/*********************************************/
/* Program solves a labyrinth.               */
/* file: Labyrinth.ino                       */
/* Author: Th. Kaffka, Cologne, Germany      */
/* Date: 05/18/2016                          */
/*********************************************/

#include <Wire.h>
#include <C:\_Thomas\Ae\C\Roboter\Interface\LCDDisplay.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\Led.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\sound.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\TouchSensor.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\GearMotor.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\Vehicle.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\CompassSensor.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\PhotoTransistor.h> // adjust this path.
#include <C:\_Thomas\Ae\C\Roboter\Interface\TouchSensor.h> // adjust this path.

const int SOUND_PIN = 11;

const int PMWA = 5;
const int AIN1 = 8;
const int AIN2 = 7;
const int ENCA = 2;

const int PMWB = 6;
const int BIN1 = 9;
const int BIN2 = 10;
const int ENCB = 3;

const int COLOR_WALL = 0;
const int COLOR_AREA = 1;
const int TURN_LEFT = -1;
const int TURN_RIGHT = 1;

// symbols.
const byte NO_DATA = 0;
const byte EXIT_FOUND = 1;
const byte NO_EXIT_FOUND = 2;
const byte WALL = 3;
const byte TIME_OVER = 4;
const byte MOVE_UP = 5;
const byte MOVE_BACK_LONG = 6;
const byte MOVE_BACK_SHORT = 7;
const byte MOVE_LEFT = 8;
const byte MOVE_RIGHT = 9;
const byte BOT_STOP = 10;
const byte BOT_UP = 11;
const byte BOT_BACK_LONG = 12;
const byte BOT_BACK_SHORT = 13;
const byte BOT_LEFT = 14;
const byte BOT_RIGHT = 15;

// parameter.
const int STACK_LEN       = 10; // the length of the stack.
const int SPEED_CALIB     = 25; // motor speed for calibration.
const int SPEED_LAB       = 40; // motor speed for labyrinth.
const int SPEED_ROT       = 35; // motor speed for rotation.
const int LIGHT_INIT_CORR = 3;  // correction for light initialization.
const int LIGHT_CORR      = 6;  // correction for light intesity.
const int LIGHT_AMB_CORR  = 2;  // correction for light ambient.
const int TIMEOUT_BORDER  = 35; // border for time out.
const int COMPASS_CORR    = 5;  // correction of the compass direction controlling.

GearMotor motor_a(PMWA, AIN1, AIN2, ENCA); 
GearMotor motor_b(PMWB, BIN1, BIN2, ENCB);
Vehicle vehicle(4.2f, 17.5f, &motor_a, &motor_b);

LCDDisplay disp = LCDDisplay();
TouchSensor touch = TouchSensor(12);
Led led = Led(13);
CompassSensor comp = CompassSensor();
PhotoTransistor ptran = PhotoTransistor(0);
TouchSensor bumper = TouchSensor(15);

boolean start;
boolean led_on;
float comp_direction[4];
int light_intensity[3]; // 0 = light max, 1 = light min, 2 = light mid.
byte goal_stack[STACK_LEN];
byte top_stack = -1;
boolean maze_end = false;
int compass_ndx = 0;
int light_plus = 0;

void setup() {
  Wire.begin();
  disp.initLCD();
  disp.textOutLCD("############", "############", "############");
  delay(1000);
  disp.clearLCD();
  delay(1000);
  disp.textLineOutLCD("Labyrinth", 1, 0);

  start = true;
  led_on = false;

  initDirection();
  initLineSensor();
  disp.writeIntegerLCD(light_intensity[0], 10, 2, 0);
  disp.writeIntegerLCD(light_intensity[1], 10, 2, 4);
  disp.writeIntegerLCD(light_intensity[2], 10, 2, 8);

  beepMessageReady(SOUND_PIN);
}

void loop() {
  // Bot start/stop statement.
  if (start && touch.isTouch()) {
    start = false;
    led.ledOff();
    beepMessageOk(SOUND_PIN);
    delay(3000);
    return;
  } else if (start) {
    led_on = !led_on;
    if (led_on) {
      led.ledOn();
    } else {
      led.ledOff();
    }
    delay(500);
    return;
  }

  pushStack(MOVE_UP);
  while (!maze_end) {
    ruleBasis();
  }

  // move the robot manual to the start position.
  delay(2000);
  compass_ndx = 0;
  top_stack = -1;
  maze_end = false;
  start = true;
  light_plus = 0;
  beepMessageReady(SOUND_PIN);
}

// initializes the main directions.
void initDirection() {
  comp_direction[0] = comp.getCompassData3599();
  comp_direction[1] = corrDirection(comp_direction[0] + 90);
  comp_direction[2] = corrDirection(comp_direction[0] + 180);
  comp_direction[3] = corrDirection(comp_direction[0] + 270);
}

// corrects the direction, if overflow.
float corrDirection(float dir) {
  if (dir > 359) dir -= 360;
  else if (dir < 0) dir += 360;
  return dir;
}

// get an average value of the light intensity.
int getLight(int x) {
   int light = 0;
   for (int i=0;i<x;i++) {
     light += ptran.readValue();
     delay(10);
   }
   return light / x;
}

// initializes the line sensor.
void initLineSensor() {
  vehicle.stop();
  vehicle.setSpeed(SPEED_CALIB);

  // calibrate the bright light.
  light_intensity[0] = getLight(10);
  
  // calibrate the low light.
  int light_value = getLight(3);
  int light_value_next = -1000;
  int light_min = 1000;
  vehicle.forward(0);
  while(light_value > light_value_next && !touch.isTouch()) {
    light_value = getLight(3);
    if (light_min > light_value) light_min = light_value;
    delay(10);
    light_value_next = getLight(3) - LIGHT_INIT_CORR;
  }
  light_intensity[1] = light_min;
  // calculate the mid. light intensity.
  light_intensity[2] = light_intensity[1] + ((light_intensity[0] - light_intensity[1]) / 2);

  // go to the start position.
  vehicle.stop();
  delay(1000);
  vehicle.setSpeed(SPEED_LAB);
  vehicle.backward(7);
}

// tests, if the linesensor is on a wall.
int getLightStatus() {
  int light = getLight(3);
  if (light < (light_intensity[2] + light_plus) - LIGHT_CORR)
    return COLOR_WALL;
  return COLOR_AREA;
}

// get the ambient light.
int getAmbientLight() {
  int light = getLight(3);
  light_plus = 0;
  if (light > light_intensity[0] + LIGHT_AMB_CORR) {
    light_plus = light - light_intensity[0];
  }
}

// push an item onto the stack.
void pushStack(byte item) {
  top_stack++;
  if (top_stack == STACK_LEN) top_stack = STACK_LEN-1;
  goal_stack[top_stack] = item;
}

// pop an item from top of stack.
byte popStack() {
  if (top_stack == -1) return -1;
  byte item = goal_stack[top_stack];
  top_stack--;
  if (top_stack < 0) top_stack = -1;
  return item;
}

// the artificial intelligence of the robot.
void ruleBasis() {
  byte stack_value = goal_stack[top_stack];
  popStack();
  if (stack_value == EXIT_FOUND) {
    move_engine(BOT_STOP);
    maze_end = true;
    beepMessageEnd(SOUND_PIN);
  } else if (stack_value == NO_EXIT_FOUND) {
    move_engine(BOT_STOP);
    maze_end = true;
    beepMessageError(SOUND_PIN);
  } else if (stack_value == WALL) {
    pushStack(MOVE_UP);
    pushStack(MOVE_BACK_SHORT);
    pushStack(MOVE_RIGHT);
    pushStack(MOVE_BACK_LONG);
  } else if (stack_value == TIME_OVER) {
    pushStack(MOVE_UP);
    pushStack(MOVE_BACK_LONG);
    pushStack(MOVE_LEFT);
    pushStack(MOVE_BACK_SHORT);
  } else if (stack_value == MOVE_UP) {
    move_engine(BOT_UP);
  } else if (stack_value == MOVE_BACK_LONG) {
    move_engine(BOT_BACK_LONG);
  } else if (stack_value == MOVE_BACK_SHORT) {
    move_engine(BOT_BACK_SHORT);
  } else if (stack_value == MOVE_LEFT) {
    move_engine(BOT_LEFT);
  } else if (stack_value == MOVE_RIGHT) {
    move_engine(BOT_RIGHT);
  } else if (stack_value == NO_DATA) {
    top_stack = -1;
  }
}

// moves the robot.
void move_engine(byte command) {
  disp.textLineOutLCD("             ", 3, 0);
  if (command == BOT_STOP) {
    disp.textLineOutLCD("stop", 3, 0);
    vehicle.stop();
  } else if (command == BOT_UP) {
    disp.textLineOutLCD("up", 3, 0);
    pushStack(move_since_wall());
  } else if (command == BOT_BACK_LONG) {
    disp.textLineOutLCD("back long", 3, 0);
    move_back(6);
    getAmbientLight();
  } else if (command == BOT_BACK_SHORT) {
    disp.textLineOutLCD("back short", 3, 0);
    move_back(4);
    getAmbientLight();
  } else if (command == BOT_LEFT) {
    disp.textLineOutLCD("left", 3, 0);
    rotate(TURN_LEFT);
  } else if (command == BOT_RIGHT) {
    disp.textLineOutLCD("right", 3, 0);
    rotate(TURN_RIGHT);
  }
}

// move since a wall is reached, or a time out occurs.
byte move_since_wall() {
  resetEncoderCountA();
  vehicle.setSpeed(SPEED_LAB);
  vehicle.forward(0);
  byte result = NO_DATA;
  while (result == NO_DATA && !maze_end) {
    if (touch.isTouch()) maze_end = true;
    if (bumper.isTouch()) {
      result = EXIT_FOUND;
    } else if (getLightStatus() == COLOR_WALL) {
      result = WALL;
    } else if (getEncoderCountA() >= TIMEOUT_BORDER) {
      result = TIME_OVER;
    }
    delay(40);
  }
  vehicle.stop();
  return result;
}

// moves the robot back.
void move_back(int len) {
  vehicle.setSpeed(SPEED_LAB);
  vehicle.backward(len);
}

// rotate to the next compass direction.
void rotate(int dir) {
  vehicle.setSpeed(SPEED_ROT);
  if (dir == TURN_RIGHT) {
    compass_ndx++;
    if (compass_ndx > 3) compass_ndx = 0;
  } else if (dir == TURN_LEFT) {
    compass_ndx--;
    if (compass_ndx < 0) compass_ndx = 3;
  }
  int direction_low = comp_direction[compass_ndx] - COMPASS_CORR;
  int direction_high = comp_direction[compass_ndx] + COMPASS_CORR;
  float compass_dir = comp.getCompassData3599();
  vehicle.rotateOnly(dir);
  while (!(compass_dir >= direction_low 
        && compass_dir <= direction_high) && !maze_end) {
    if (touch.isTouch()) maze_end = true;
    compass_dir = comp.getCompassData3599();
  }
  vehicle.stop();
}
